#
# rcsys.rb
#
#   Copyright (c) 2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   you can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'est/confp'
require 'amstd/must'


module EST

  class ConfigError < StandardError; end


  class ObjectDB

# CORE.odb.add 'sender', FocusingCluster.new
# CORE.odb.add 'popper', FocusingCluster.new
# CORE.odb.add 'getter', AgentCluster.new
# CORE.odb.add 'mailbox', Mailset.new
# CORE.odb.add 'composer', ObjectCluster.new
# CORE.odb.add 'refiler', FocusingCluster.new

    def initialize
      @obj = {}
    end

    def add( name, obj )
      @obj[ name ] = obj
    end

    def get( name )
      unless ret = @obj[ name ] then
        raise ArgumentError, "object cluster #{name} does not exist"
      end
      ret
    end

    def read( fn )
      ent = ConfigFileParser.parse( fn )
      FormatCheck.check ent, <<End
{
  sender:h;
  popper:h;
  mailbox:h;
  composer:h;
  refiler:h;
  getter:h;
  profile:h;
  interface:{
    #{EST::Interface}:h;
  }
  first-profile:s;
}
End

      fp = ent.delete( 'first-profile' )
      ent.each do |nm, rc|
        if nm == 'interface' then
          rc[ EST::Interface ].each do |k,v|
            get( k ).map v
          end
        else
          get( nm ).map rc
        end
      end

      fp
    end

    def write_back
    end

  end



  class ObjectRc
  
    def initialize( file, path, dat )
      @file = file
      @path = path
      @dat  = dat
    end

    attr :path

    def []( key )
        #unless ret = @dat[key] then
        #  raise ConfigError, "resource '#{key}' does not exist"
        #end
        #ret
      @dat[ key ]
    end

    def []=( key, val )
      @dat[ key ] = val
    end

    def key?( key )
      @dat.key? key
    end

    def to_hash
      @dat.dup
    end

  end


  class ArrayRcNode
  
    def initialize( dat )
      @name = nm
      @dat = dat
    end

    def []( i )
      unless ret = @dat[i] then
        raise ConfigError, "resource '#{nm}[#{i}]' does not exist"
      end
      ret
    end

    def []=( i, dat )
      if @dat.size >= i then
        raise ConfigError, "resource '#{@name}[#{i}]' out of range"
      end
      @dat[i] = dat
    end

    def each( &block )
      @dat.each( &block )
    end

    def key?( i )
      @dat[i] ? true : false
    end

    def size
      @dat.size
    end

    def to_a
      @dat.dup
    end
  
  end



  class ObjectCluster
  
    include FormatCheck

def dprint
  puts "---- #{type} ----"
  puts '---- klass'
  @obj.each {|name,i| puts "#{name}: #{i}" }
  puts '---- object'
  @obj.each {|name,i| puts "#{name}: #{i.type}" }
end
    
    def initialize( fmt )
      @format = fmt
      @klass = {}
      @obj = {}
    end

    def map( ent )
      ent.each do |name, rc|
        add name, rc
        fmt_check rc, @format
      end
    end

    def register_class( name, klass )
      klass.must Class
      @klass[ name ] = klass
    end

def parg( name, rc )
  puts "name  ='#{name}'"
  puts "rc    =#{rc.inspect}"
  puts "@klass=#{@klass.inspect}"
end
    
    def get_class( name, rc )
      c = nil
begin
      c =
      (rc['type'] && @klass[ rc['type'] ]) ||
      (name && @klass[name]) ||
      @klass['default']
ensure
  unless c
    parg name, rc
    bug!
  end
end
      c
    end

    def createobj( name, rc, *args )
begin
      c = get_class( name, rc )
      c.new( rc, *args )
rescue ArgumentError
  p c
  raise
end
    end
    private :createobj

    def create( rc, *args )
      createobj nil, rc, *args
    end
    private :create

    def add( name, rc, *args )
      @obj[ name ] = createobj( name, rc, *args )
    end


    def get( name )
      @obj[ name ]
    end

    def each_object( &block )
      @obj.each {|k,v| yield v }
    end
    alias each each_object

  end


  class FocusingCluster < ObjectCluster

    def select( name )
      @current = get( name )
    end

    def get( name = nil )
      if name then
        super
      else
        @current
      end
    end

    attr :current

  end


  class AgentCluster

    def get( name )
      super.dup
    end
  
  end

end
