#
# mbox.rb
#
#   Copyright (c) 1999,2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
#
#   This program is free software.
#   you can distribute/modify this program under the terms of
#   the GNU Lesser General Public License version 2 or later.
#

require 'amstd/errutil'


module EST


  class Mailset

    def initialize
      super
      @flags = {}

      @init0 = @init1 = false
    end

    def init0
      return if @init0
      @init0 = true

      construct_gui
    end

    def init1
      return if @init1
      @init1 = true

      @current = inbox
      @win.show_all
      @current.select 0
    end

    def shutdown
      super
      @win.destroy
    end


    def write( str )
      @notice.set str.to_s
      @notice.queue_draw
    end


    def chmbox( newname )
      curname = @current.name
      return if curname == newname
      return unless new = self[ newname ]

      # keep mbox single select mode
      if @current.multiselect then
        @excheck.clicked
      end

      old = @current
      @current = new

      curi = @flags[ curname ]
      newi = @flags[ newname ]
      bug! unless curi and newi

      newi.viewer = @viewer
      curi.viewer = nil

      @clistscr.set_adjustment newi.listadj
      newi.clist.show_all
      curi.clist.hide_all

      if @current.size == 0 then
        CORE.notice "no mail"
      else
        CORE.notice "#{@current.size} mails"
      end
      
      @current.select 0
      @current.select 0   # this is not type miss
    end


    def discard_mail( mbox, idx )
      if mbox.name == 'trash' then
        CORE.rmmail( trash, idx )
      else
        CORE.mvmail( mbox.name, trash, idx )
      end
      if idx >= mbox.size then
        idx = mbox.size - 1
      end
      mbox.select idx
    end


    private


    #def nombox( name )


    def make_real_mbox( name, rc, lo )
      adj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      cli = init_clist
      @clistbox.pack_start( cli, true, true, 0 )
      cli.realize
      cli.set_vadjustment adj
      cli.selection_mode = Gtk::SELECTION_BROWSE
      
      @flags[ name ] = int = InternalField.new( cli, adj )
      mb = Mailbox.new( name, rc, lo, int )
      if @flags.size == 1 then
        @clistscr.set_adjustment adj
        int.mbox    = mb
        int.viewer  = @viewer
      end

      mb
    end

    def init_clist
      rc = CORE.ifrc['list']['entries']

      arr = rc.collect {|ent| ent['label'] }
      cli = Gtk::CList.new( arr )

      width = 640
      height = 170
      cli.set_usize( width, height )

      total = 0
      i = s = 0
      arr = rc.collect {|ent|
        i = (ent['rsize'] || 30).to_i
        total += i
        i
      }
      arr.each_with_index do |rel, idx|
        cli.set_column_width( idx, width * rel / total )
      end

      cli
    end


    def construct_gui
      @win = Gtk::Window.new( Gtk::WINDOW_TOPLEVEL )
      @win.set_title( 'mailest' )
      @win.border_width(0)
      @win.set_usize( 640, 480 )

      @bigbox   = Gtk::VBox.new( false, 0 )
      @win.add @bigbox

      @topbox   = Gtk::HBox.new( false, 0 )
      @bigbox.pack_start( @topbox, false, false, 0 )

      @paned = Gtk::VPaned.new
      @paned.handle_size = 7
      @paned.gutter_size = 7
      @bigbox.pack_start @paned, true, true, 0

      @excheck = Gtk::CheckButton.new('')
      @topbox.pack_start @excheck, false, false, 0

      @mboxcombo = Gtk::Combo.new
      @mboxcombo.set_usize 100, 24
      @mboxcombo.entry.set_editable false
      @mboxcombo.set_popdown_strings mbox_names()
      @mboxcombo.entry.set_text 'inbox'
      @topbox.pack_start @mboxcombo, false, false, 0

      construct_menu

      @listbox  = Gtk::HBox.new( false, 0 )
      @paned.add1 @listbox

      @clistbox = Gtk::VBox.new( false, 0 )
      @clistscr = Gtk::VScrollbar.new( Gtk::Adjustment.new(0,0,0,0,0,0) )

      @textbox = Gtk::VBox.new( false, 0 )
      @paned.add2 @textbox

      @notice = Gtk::Label.new( EST::Copyright )
      @notice.set_usize( 640, 20 )
      @notice.set_alignment( 0, 0.5 )
      @textbox.pack_start @notice, false, false, 0

      @textadj = Gtk::Adjustment.new( 0, 0, 0, 0, 0, 0 )
      @textscr = Gtk::VScrollbar.new( @textadj )
    @listbox.pack_start @clistbox, true, true, 0
    @listbox.pack_start @clistscr, false, false, 0
    @listbox.pack_start @textscr, false, false, 0
      text     = Gtk::Text.new( nil, @textadj )  # Gtk::Buffer.new
      @textbox.pack_start text, true, true, 0

      @viewer = Viewer.new( text, @textadj )

      bind_menu_key
      set_callback
    end


    def set_callback
      @win.signal_connect( 'destroy' ) do |w|
        CORE.shutdown
        false
      end

      mbox_names.each do |mbname|
        mitem = Gtk::MenuItem.new( mbname )
        mitem.signal_connect( 'activate' ) do |w|
          protect( 'move to' ){
            @current.each_selection do |idx|
              CORE.mvmail @current, mbname, idx
            end
          }
        end
        @menumove.add mitem
      end

      @excheck.signal_connect( 'toggled' ) do |w|
        protect( 'change selection mode' ) {
          @current.multiselect = ! @current.multiselect
        }
      end

      bind_menu_action

      @combo_pressing = false

      @mboxcombo.entry.signal_connect( 'changed' ) do |w|
        protect( 'combo change' ){
          unless @combo_pressing then
            chmbox w.get_text
          end
          false
        }
      end

      @mboxcombo.button.signal_connect( 'pressed' ) do |w|
        protect( 'combo pressed' ){
          @combo_pressing = true
          false
        }
      end

      @mboxcombo.button.signal_connect( 'released' ) do |w|
        protect( 'combo released' ){
          @combo_pressing = false
          tmp = @mboxcombo.entry.get_text
          chmbox tmp
          false
        }
      end
    end

  end   # class Mailset

  require 'menu'   # tmp


  class InternalField

    def initialize( clist, adj )
      @clist = clist
      @listadj = adj

      @mbox = nil
      @viewer = nil
    end

    attr :clist
    attr :listadj

    attr :mbox, true
    attr :viewer, true

  end



  class Mailbox

    include Enumerable

    def initialize( name, rc, lo, intern )
      @name   = name.dup.freeze
      @rc     = rc
      @loader = lo

      @array  = []
      @deleting = 0

      @intern = intern
      @win = intern.clist
      @adj = intern.listadj

      @clistsize = 0
      @selected = nil
      @unselected = false
      @multisel = false
      @no_view = 0

      set_signal_handler
      load
    end


    attr :name

    attr :selected
    attr :unselected

    def size
      @array.size
    end
    alias length size

    def empty?
      @array.empty?
    end


    def []( idx )
      return nil unless idx
      @array[idx]
    end
    
    def each( &block )
      @array.each( &block )
    end

    def each_index( &block )
      @array.each_index( &block )
    end

    def each_mail
      @array.each {|port| yield Mail.new( port ) }
    end

    def index( port )
      port.must ::TMail::Port if $DEBUG
      @array.index port
    end

    def rindex( port )
      port.must ::TMail::Port if $DEBUG
      @array.rindex port
    end

    def indexes( *args )
      @array.indexes *args
    end

    def to_a
      @array.compact
    end


    private


    def load( no_cache = false )
      updating { do_load no_cache }
    end

    def cache_signature
      ''
    end


    public


    def reload
      updating {
        @array.each {|port| port.ref}
        @array.clear
        do_load true
      }
      select 0
    end

    def update
dputs 'upd'
      updating {
        hash = {}
        @array.each {|port| hash[ port ] = true }
        ur = (@name == 'inbox')
        @loader.each do |p|
          p.unread = ur
          add p unless hash.key? p
        end
      }
    end


    def new_port
      @loader.new_port
    end
    alias new new_port

    def add( port )
      port.ref
      @array.push port

      c = nil
      s = ready_setter
      @win.append [''] * s.size

      begin
        if c = port.cache then
          s.each_index do |i|
            s[i].setv( @win, size - 1, i, c[i] )
          end
        else
          c = []
          mail = Mail.new( port )
          s.each_index do |i|
            c[i] = s[i].set( @win, size - 1, i, mail )
          end
          port.cache = c
        end
      rescue
        if port.cache then
          port.cache = nil
          retry
        else
          raise
        end
      end
    end

    def move_of( from, idx )
      p = @loader.new_port
      port = from.moveof_rm( idx )
dputs "do_mv: from=#{port},to=#{p}"
      port.mv p
      p.unread = port.unread
      p.cache = port.cache
      add p

      p
    end

    def moveof_rm( idx )
      del idx
    end
    protected :moveof_rm

    def remove( idx )
      port = del( idx ); return nil unless port
      port.unref
    end

    def delete_if
      updating {
        deleting {
          @array.each_with_index do |port,idx|
            if yield( port ) then
              remove idx
            end
          end
        }
      }
    end

    def move_if
      updating {
        deleting {
          @array.each_with_index do |port,idx|
            if mb = yield( port ) then
              mb.move_of self, idx
            end
          end
        }
      }
    end


    private


    def del( idx )
      if @deleting > 0 then
        port = @array[idx]
        @array[idx] = nil
      else
dputs "remove row #{idx}"
        port = @array.delete_at( idx )
        @win.remove_row idx
      end
      port
    end

    def deleting
      begin
        @deleting += 1
        yield
      ensure
        @deleting -= 1
        if @deleting == 0 then
dputs 'compact'
          if not @array.empty? and @array.nitems < @array.size then
            (@array.size - 1).downto(0) do |idx|
              @win.remove_row idx unless @array[idx]
            end
          end
          @array = ret.compact
        end
      end
    end

    # stop redraw
    def updating
      begin
        @no_view += 1
        if li = @intern.clist then
          li.freeze
        end
        yield
      ensure
        li.thaw if li
        @no_view -= 1
      end
    end


    public


    def set_signal_handler
      @win.signal_connect( 'select_row' ) do |w,row,col,eve|
      protect( 'select' ) {
        @selected = row
        CORE.mailbox.write ''
        adjust_to row
        view
      }
      end

      @win.signal_connect( 'unselect_row' ) do |w,row,col,eve|
      protect( 'unselect' ) {
        @unselected = true
      }
      end
@s = nil
    end
    private :set_signal_handler

    def select( idx )
      return unless legal_index? idx
      @win.select_row idx, 0
    end

    def set_select_handler
    end

    def legal_index?( idx )
      idx and idx >= 0 and idx < size
    end

    def each_selection
      @win.each_selection{|idx| yield idx }
    end

    #
    # select and view
    #

    def view_next
      select @selected + 1
    end

    def view_prev
      select @selected - 1
    end

    #
    # select column only
    #
    
    def select_next
      no_view {
        select @selected + 1
      }
    end

    def select_prev
      no_view {
        select @selected - 1
      }
    end

    def sync_view
      if @selected != @viewing then
        select @selected
      end
    end

    
    def no_view
      begin
        @no_view += 1
        yield
      ensure
        @no_view -= 1
      end
    end

    def view
      return unless v = @intern.viewer
      return if @no_view > 0
      @viewing = @selected
      port = self[ @viewing ]
      port.unread = false
  ####### pending                   @intern.list.redraw @viewing
      v.view port
    end

    def adjust_to( idx )
      return if idx < 0 or idx >= size

      adj = @adj
      val = adj.value.to_i
      page = adj.page_size.to_i
      row = @win.row_height + 1   # 1 for spacing
      v = row * (idx + 3)
      if v > val + page then
# puts 'v'
        adj.value = ret = [v - page, adj.upper.to_i - page].min
      else
        v = row * (idx - 1)
        if v < val then
# puts '^'
          adj.value = ret = [v, adj.lower].max
        end
      end
# puts "adj.value=#{val} set=#{ret} row=#{row} page=#{page}"
    end
    private :adjust_to


    def multiselect
      @multisel
    end

    def multiselect=( arg )
      @multisel = arg
      if @multisel then
        @no_view += 1
        @win.selection_mode = Gtk::SELECTION_EXTENDED
      else
        @no_view -= 1
        @win.selection_mode = Gtk::SELECTION_BROWSE
      end
    end

    
    def repaint
      @win.freeze
      each_index do |i|
        listmail i
      end
      @win.thaw

      select 0
    end
    

    def getter
      CORE.ifrc['list']['entries'].collect do |ent|
        g = CORE.getter.get( ent['getter'] ).dup
        unless g then
          bug! "is nil: #{ent.inspect}"
        end
        g
      end
    end
      
    def setter
      CORE.ifrc['list']['entries'].collect do |ent|
        s = CORE.setter.get( ent['setter'] ).dup
        s
      end
    end

    def ready_setter
      return @s if @s   # tmp

      s = setter
      g = getter
bug! 'set' if s.size > s.nitems
bug! 'get' if g.size > g.nitems
      s.each_with_index do |st, i|
        st.ready( g[i], @win )
      end
      @s = s
      s
    end

  end   # class Mailbox


end   # module EST
